/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/app.cpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Application class implementation.
**/

#include "app.hpp"

#include <iostream>
#include <iomanip>

#include <argp.h>

/**
    Program version string. It is printed in respond to `--version` option. The string may include
    copyright.
**/
char const * argp_program_version =
    PACKAGE_STRING "\n"
    "Copyright (C) " PACKAGE_COPYRIGHT_YEAR " " PACKAGE_COPYRIGHT_HOLDER ".\n"
    "License " PACKAGE_LICENSE " <" PACKAGE_LICENSE_URL ">.\n"
    "This is free software: you are free to change and redistribute it." "\n"
    "There is NO WARRANTY, to the extent permitted by law."
    ;

/** Email address to send bug reports to. May be `nullptr`. **/
char const * argp_program_bug_address = PACKAGE_BUGREPORT;

/** Status to return in case of command line error. **/
error_t argp_err_exit_status = 2;

// -------------------------------------------------------------------------------------------------
// app_t
// -------------------------------------------------------------------------------------------------

app_t::app_t(
    int     argc,
    char *  argv[]
) :
    OBJECT_T(),
    prog_name( program_invocation_short_name ),
    mode( default_mode() )
{
    TRACE();
    posix::signal::set_t set;
    set += SIGINT;
    set += SIGTERM;
    posix::signal::mask( SIG_BLOCK, set );
    parse_argv( argc, argv );
    INF( PACKAGE_STRING "." );
    INF(
        PACKAGE_NAME " is a keyboard layout selector for Gnome Shell and X Window System: "
        "it selects specified keyboard layout when corresponding key is tapped."
    );
    INF( "Starting in " << mode << " mode." );
    init();
    layout_switcher_t switcher( * kb, map, mode, bell );
    switcher.start();
    show_map();
    INF( "Press Ctrl+C to exit..." );
    posix::signal::wait( set );
    switcher.stop();
    INF( "Done." );
}; // ctor

app_t::mode_t
app_t::default_mode(
) {
    mode_t mode = mode_t::sm_xkb;
    #if ENABLE_GNOME
        auto var = posix::get_env( "DESKTOP_SESSION" );
        if ( var == "gnome" or var == "gnome-xorg" ) {
            mode = mode_t::sm_gnome;
        }; // if
    #endif
    return mode;
}; // default_mode

/**
    Parses comman line option or argument. The method is static because it is called from whithin
    `argp_parse`.
**/
::error_t
app_t::parse_opt_or_arg(
    int                 key,        ///< Option key (a short name or integer id).
    char *              _arg,       ///< String to parse.
    struct argp_state * state
) {

    char const * arg = _arg;
        //  It is a little trick. We are not going to change argument, but parameter
        //  should be of `char *` type.

    app_t * app = reinterpret_cast< app_t * >( state->input );

    try {       // No exceptions should leak from this method.

        switch ( key ) {

            case ok_list_keys : {
                app->list_keys();
                exit( EXIT_SUCCESS );
            } break;

            case ok_list_layouts : {
                app->list_layouts();
                exit( EXIT_SUCCESS );
            } break;

            case ok_xkb : {
                if ( state->next - 1 > 1 ) {
                    ERR1( app, "Must be the first option in the command line." );
                }; // if
                app->mode = mode_t::sm_xkb;
            } break;

            case ok_gnome : {
                if ( ENABLE_GNOME ) {
                    if ( state->next - 1 > 1 ) {
                        ERR1( app, "Must be the first option in the command line." );
                    }; // if
                    app->mode = mode_t::sm_gnome;
                } else {
                    ERR1( app, "Program is built without Gnome support." );
                }; // if
            } break;

            case ok_bell : {
                app->bell = true;
            } break;

            case ARGP_KEY_NO_ARGS : {
                arg = "LFSH=1";
                app->parse_arg( arg );
                arg = "RTSH=2";
                app->parse_arg( arg );
            } break;

            case ARGP_KEY_ARG : {
                app->parse_arg( arg );
            } break;

            case ARGP_KEY_END : {
            } break;

            default : {
                return ARGP_ERR_UNKNOWN;
            } break;

        }; // switch

    } catch ( std::exception const & ex ) {

        char const * format = nullptr;
        switch ( key ) {
            case ARGP_KEY_ARG : {
                format = "Bad argument `%s': %s";
            } break;
            case ARGP_KEY_NO_ARGS : {
                format =
                    "Oops, bad default argument `%s': %s; "
                    "try to specify explicit argument(s) to avoid this problem";
            } break;
            default : {
                format = "Bad option `%s': %s";
                arg = state->argv[ state->next - 1 ];
            } break;
        }; // switch
        argp_error( state, format, arg, ex.what() );

    }; // try

    return 0;

}; // parse_opt_or_arg

/** Parses entire command line. **/
void
app_t::parse_argv(
    int     argc,
    char *  argv[]
) {

    argp_option opts[] = {
        { "Mode options:",  0,                  nullptr, OPTION_DOC | OPTION_NO_USAGE,
            "",
            100 },
        { "xkb",            ok_xkb,             nullptr, 0,
            "Run in Xkb mode",
            101 },
        { "gnome",      ok_gnome,               nullptr, 0,
            "Run in Gnome mode",
            101 },

        { "Other options:", 0,                  nullptr, OPTION_DOC | OPTION_NO_USAGE,
            "",
            200 },
        { "bell",           ok_bell,            nullptr, 0,
            "Ring a bell when a layout is selected",
            201 },

        { "Help options:",  0,                  nullptr, OPTION_DOC | OPTION_NO_USAGE,
            "",
            -2  },
        { "list-keys",      ok_list_keys,       nullptr, 0,
            "List key codes and names",
            -1  },
        { "list-layouts",   ok_list_layouts,    nullptr, 0,
            "List layout numbers and names",
            -1  },

        { 0 }

    }; // opts

    string_t const note =
    #if ENABLE_GNOME
        ""
    #else
        "Note: The program is built without Gnome support, "
        "so `--gnome' option is not available."
        "\n\n"
    #endif
    ;

    string_t const doc = STR(
        "\n"
        PACKAGE_NAME " is a keyboard layout selector for X Window System: "
        "it selects specified keyboard LAYOUT when corresponding KEY is tapped."
        "\v"    // Options will be listed here.
        "KEY is either a key code (unsigned integer) "
        "or a symbolic key name (`LFSH', `RALT', etc). "
        "Run `" << prog_name << " --list-keys' "
        "to see the list of keys currently known to the X Window server. "
        "\n\n"
        "\"Tap\" means that a single key is pressed and quickly released. "
        "If a key is released slowly, "
        "or pressed/released in combination with other key(s), "
        "it is not a tap. "
        "\n\n"
        "LAYOUT depends on the mode. "
        "In Xkb mode LAYOUT is an Xkb \"group\". "
        "In Gnome mode LAYOUT is a Gnome \"input source\". "
        "In any case, LAYOUT is either a group/input source number (starting from 1) "
        "LAYOUT is an Xkb \"group\", "
        "or its name. "
        "Run `" << prog_name << " --list-layouts' "
        "to see the list of currently configured layouts. "
        "\n\n"
        "Mode is selected by options `--xkb' and `--gnome'. "
        "These options are mutually exclusive. "
        "If mode is not specified explicitly " PACKAGE_NAME " "
        "looks to environment, if variable DESKTOP_SESSION equals \"gnome\" "
        PACKAGE_NAME " starts in Gnome mode, "
        "otherwise it starts in Xkb mode. "
        "\n\n"
        << note <<
        "If no arguments specified, `LFSH=1' and `RTSH=2' are used by default."
        "\n\n"
        "Exit status:\n"
        "  0 -- Everything is ok.\n"
        "  " << argp_err_exit_status << " -- Usage (command-line) error.\n"
        "  255 -- Runtime error.\n"
        "\n"
        "For more detaled documenatation, run `man " PACKAGE "'.\n"
        PACKAGE_NAME " home page: " PACKAGE_URL "."
    );

    argp parser = { opts, parse_opt_or_arg, "[KEY=LAYOUT...]", doc.c_str() };
    ::error_t error = argp_parse( & parser, argc, argv, 0, 0, this );
    if ( error ) {
        exit( argp_err_exit_status );
    }; // if

}; // parse_argv

/**
    Parses one argument. All arguments should follow "KEY=LAYOUT" format, where both KEY and LAYOUT
    can be either name or number of key and layout respectively.
**/
void
app_t::parse_arg(
    string_t const & arg
) {
    strings_t parts  = split( '=', arg, 2 );
    uint_t    key    = parse_key( parts[ 0 ] );
    uint_t    layout = parse_layout( parts.size() > 1 ? parts[ 1 ] : "" );
    DBG( "key " << key_name( key ) << " => layout " << layout_name( layout ) );
    // Check if the key is already assigned.
    auto it = map.find( key );
    if ( it != map.end() ) {
        ERR( "Key " << key_name( key ) << " is already assigned." );
    }; // if
    map[ key ] = layout;
}; // parse_arg

uint_t
app_t::parse_key(
    string_t const & key
) {
    // TODO: Allow key in form "50:LFSH".
    uint_t code = 0;
    if ( key.empty() ) {
        ERR( "No key specified." );
    }; // if
    if ( like_integer( key ) ) { // Looks like a key code.
        string_t prefix = "Bad key code: ";
        code = parse_int( key, prefix );
        key_code_range_t range = key_code_range();
        if ( not ( range.min <= code and code <= range.max ) ) {
            ERR( prefix << "Out of range [" << range.min << ".." << range.max << "]." );
        }; // if
    } else {                       // Looks like a key name.
        string_t name = key;
        if ( left( name, 1 ) == "<" and right( name, 1 ) == ">" ) {
            name = name.substr( 1, name.size() - 2 );
        }; // if
        auto names = key_names();
        auto pos   = find( name, names );
        if ( pos == npos ) {
            // If not found, try to find upper-case variant of key name.
            pos = find( uc( name ), names );
        }; // if
        if ( pos == npos ) {
            ERR( "Bad key name: Unknown name." );
        }; // if
        code = pos;
    }; // if
    return code;
}; // parse_key

uint_t
app_t::parse_layout(
    string_t const & layout
) {
    uint_t code = 0;
    if ( layout.empty() ) {
        ERR( "No layout specified." );
    }; // if
    if ( like_integer( layout ) ) {
        string_t prefix = "Bad layout number: ";
        code = parse_int( layout, prefix );
        uint_t min = 1;
        uint_t max = 0;
        switch ( mode ) {
            case mode_t::sm_xkb : {
                max = x::layout_max;
            } break;
            #if ENABLE_GNOME
                case mode_t::sm_gnome : {
                    max = gnome::layout_max;
                } break;
            #endif
            default : {
                ERR( "oops" );
            } break;
        };
        if ( not ( min <= code and code <= max ) ) {
            ERR( prefix << "Out of range [" << min << ".." << max << "]." );
        }; // if
        code = code - 1;
    } else {
        auto name  = layout;
        auto names = layout_names();
        auto it    = find( name, names );
        // TODO: What if layout names are not unique?
        if ( it == npos ) {
            ERR( "Bad layout name: Unknown name." );
        }; // if
        code = it;
    }; // if
    return code;
}; // parse_layout

uint_t
app_t::parse_int(
    string_t const & str,
    string_t const & prefix
) {
    int num = 0;
    try {
        num = val< decltype( num ) >( str );
    } catch ( val_error_t const & ex ) {
        ERR( prefix << ex.what() );
    };
    if ( num < 0 ) {
        ERR( prefix << "Negative number is not allowed." );
    };
    return num;
};

void
app_t::init(
) {
    if ( not display ) {
        x::init();
        display.reset( new x::display_t );
        kb.reset( new x::kb_t( * display ) );
        x::kb_t::desc_p kb_desc( kb->desc() );
        _key_code_range = kb_desc->key_code_range();
        _key_names      = kb_desc->key_names();
        switch ( mode ) {
            case mode_t::sm_xkb : {
                _layout_names = kb_desc->group_names();
            } break;
            #if ENABLE_GNOME
                case mode_t::sm_gnome : {
                    gnome::input_source_manager_t ism;
                    _layout_names = ism.keyboard_layouts();
                } break;
            #endif // ENABLE_GNOME
            default : {
                ERR( "oops" );
            } break;
        }; // switch
    }; // if
}; // init

/** Prints list of known keys. **/
void
app_t::list_keys(
) {
    strings_t const & names = key_names();
    for ( size_t i = 0, end = names.size(); i < end; ++ i ) {
        if ( not names[ i ].empty() ) {
            std::cout << i << ":" << names[ i ] << "\n";
        }; // if
    }; // for i
}; // list_keys

/** Prints list of known layouts. **/
void
app_t::list_layouts(
) {
    strings_t const & names = layout_names();
    for ( size_t i = 0, end = names.size(); i < end; ++ i ) {
        if ( not names[ i ].empty() ) {
            std::cout
                << i + 1 << ":"
                << names[ i ]
                << "\n";
        }; // if
    }; // for i
}; // list_layouts

void
app_t::show_map(
) {
    auto keys    = key_names();
    auto layouts = layout_names();
    size_t kw    = 0;   // Max width of key name.
    // I want the listing is sorted by layout numbers, not keys.
    // I also want key names are aligned vertically.
    // So I build the layout number => key name map first:
    std::map< map_t::mapped_type, string_t > pam;
    for ( auto it = map.begin(), end = map.end(); it != end; ++ it ) {
        string_t name = key_name( it->first );
        kw = std::max( kw, name.size() );
        pam.insert( { it->second, name } );
    };
    // And now print the map:
    for ( auto it = pam.begin(), end = pam.end(); it != end; ++ it ) {
        INF(
            "Tap key " << std::setw( kw ) << it->second << " "
            << "to select layout "  << layout_name( it->first ) << "."
        );
    }; // for it
}; // show_map

app_t::key_code_range_t const &
app_t::key_code_range() {
    init();
    return _key_code_range;
}; // key_code_range

strings_t const &
app_t::key_names() {
    init();
    return _key_names;
}; // key_names

strings_t const &
app_t::layout_names() {
    init();
    return _layout_names;
}; // layout_names

string_t
app_t::key_name(
    uint_t code
) {
    auto names = key_names();
    auto name  = code < names.size() ? names[ code ] : "";
    return STR( code << ( name.empty() ? "" : ":" + name ) );
}; // key_name

string_t
app_t::layout_name(
    uint_t code
) {
    auto names = layout_names();
    auto name  = code < names.size() ? names[ code ] : "";
    return STR( code + 1 << ( name.empty() ? "" : ":" + name ) );
}; // layout_name

// end of file //
