/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/base.hpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

#ifndef _BASE_HPP_
#define _BASE_HPP_

#include "config.h"

#include <exception>
#include <memory>           // std::unique_ptr

#include <cstddef>          // size_t

#include "string.hpp"
#include "ui.hpp"

typedef unsigned int uint_t;
typedef std::size_t  size_t;
template< typename obj_t > using ptr_t = std::unique_ptr< obj_t >;

/**
    Base class for all other classes.

    Objects are not copyable.

    It provides debugging facilities: each object has type (object type is used for constructing
    object id but is not saved as data member), number, and id.

    Object type is a string, it should be specified in the constructor. It is expected `OBJECT_T()`
    is used for it:

    @code {.cpp}
        class class_t : public object_t {
            public :
                class_t() : OBJECT_T() {
                    ...
                };
                ...
        };
    @endcode

    Object number is assigned automatically, sequentially within the type. The first created object
    of the type receives no 1, the next created object of the same type receives number 2, and so
    on. Numbers of deleted objects are not reused.

    Object id is constructed from object type and object number, and looks like "object#1".

    All these stuff is used for debugging purposes solely.

**/
class object_t {

    // Objects are non-copyable.
    object_t( object_t const & ) = delete;
    object_t & operator =( object_t const & ) = delete;

    #if defined( DEBUG )

        public :

            explicit object_t( char const * type );
            virtual ~object_t();

            object_t const * _this_() const;
            uint_t           _no_()   const;
            string_t         _id_()   const;

        private :

            uint_t const    _no;    ///< Object number, unique within the type.
            string_t const  _id;    ///< Object id, unique within the program.
    #else

        public :

            explicit object_t( char const * type ) {};

    #endif

}; // object_t

#if defined( DEBUG )
    #define OBJECT_T()  object_t( __FUNCTION__ )
    object_t const * _this_();
    void _dbg( object_t const * obj, char const * file, int line, string_t const & mesg );
    #define DBG1( _OBJ_, _ARGS_ ) _dbg( _OBJ_, __FILE__, __LINE__, STR( _ARGS_ ) )
#else
    #define OBJECT_T()  object_t( nullptr )
    #define DBG1( _OBJ_, _ARGS_ ) void( 0 )
#endif

#define WRN1( _OBJ_, _ARGS_ ) {                                                                    \
    string_t _MSG_ = STR( _ARGS_ );                                                                \
    msg( sl_warning, _MSG_ );                                                                      \
}
#define ERR1( _OBJ_, _ARGS_, ... ) {                                                               \
    string_t _MSG_ = STR( _ARGS_ );                                                                \
    DBG1( _OBJ_, "Throwing: " << _MSG_ );                                                          \
    throw error_t( _MSG_, ## __VA_ARGS__ );                                                        \
}
#define DBG( _ARGS_ )       DBG1( _this_(), _ARGS_ )
#define INF( _ARGS_ )       msg( sl_info, STR( _ARGS_ ) )
#define WRN( _ARGS_ )       WRN1( _this_(), _ARGS_ )
#define ERR( _ARGS_, ... )  ERR1( _this_(), _ARGS_, ## __VA_ARGS__ )

#define CATCH_ALL1( _OBJ_, _STATEMENT_ ) {                                                         \
    try {                                                                                          \
        _STATEMENT_;                                                                               \
    } catch ( std::exception const & ex ) {                                                        \
        WRN1( _OBJ_, ex.what() );                                                                  \
    } catch ( ... ) {                                                                              \
        WRN1( _OBJ_, "Unknown exception occurred" );                                               \
    };                                                                                             \
}

#define CATCH_ALL( _STATEMENT_ ) \
    CATCH_ALL1( _this_(), _STATEMENT_ )

#if defined( DEBUG )

    class _func_tracer_t {
        public :
            _func_tracer_t( object_t const * obj, char const * file, int line, char const * func );
            ~_func_tracer_t();
        private :
            object_t const * const  _obj;
            char const * const      _file;
            int const               _line;
            char const * const      _func;
    }; // class _func_tracer_t

    #define TRACE1( _OBJ_ ) \
        _func_tracer_t _tracer_ ## __COUNTER__( _OBJ_, __FILE__, __LINE__, __FUNCTION__ )
    #define TRACE() \
        TRACE1( _this_() )
#else

    #define TRACE1( _OBJ_ )
    #define TRACE()

#endif

#endif // _BASE_HPP_

// end of file //
