/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/gnome.cpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    Gnome wrappers implementation.
**/

#include "gnome.hpp"

#include <iostream>
#include <json-glib/json-glib.h>

#define CHECK( error ) {                                                                           \
    if ( error ) {                                                                                 \
        ERR( error->message, error );                                                              \
    };                                                                                             \
}

namespace gnome {

// -------------------------------------------------------------------------------------------------
// error_t
// -------------------------------------------------------------------------------------------------

error_t::error_t(
    string_t const & message,
    GError *         error
) :
    parent_t( message ),
    _error( error )
{
}; // ctor

error_t::error_t(
    myself_t const & other
) :
    parent_t( other.what() ),
    _error( g_error_copy( other._error ) )
{
};

error_t::~error_t(
) {
    g_error_free( _error );
}; // dtor

bool
error_t::matches(
    GQuark  domain,
    gint    code
) const {
    return g_error_matches( _error, domain, code );
};

// -------------------------------------------------------------------------------------------------
// input_source_manager_t
// -------------------------------------------------------------------------------------------------

input_source_manager_t::input_source_manager_t(
) :
    OBJECT_T()
{
    _reset();
};

input_source_manager_t::~input_source_manager_t(
) {
};

void
input_source_manager_t::_reset(
) {
    GError * error = nullptr;
    _bus.reset( g_dbus_proxy_new_for_bus_sync(
        G_BUS_TYPE_SESSION,
        G_DBUS_PROXY_FLAGS_NONE,
        nullptr,
        "org.gnome.Shell",
        "/org/gnome/Shell",
        "org.gnome.Shell",
        nullptr,
        & error
    ) );
    CHECK( error );
};

string_t
input_source_manager_t::_eval(
    string_t const & jscript
) {
    GError * error = nullptr;
    static char const * const method = "Eval";
    static char const * const format = "(bs)";
    variant_p reply( g_dbus_proxy_call_sync(
        _bus.get(),
        method,
        g_variant_new( "(s)", jscript.c_str() ),
        G_DBUS_CALL_FLAGS_NONE,
        -1,
        nullptr,
        & error
    ) );
    CHECK( error );
    if ( not g_variant_check_format_string( reply.get(), format, TRUE ) ) {
        ERR(
            "Can't parse Gnome Shell response: "
            << "method `" << method << "' returned response which does not match "
            << "format `" << format << "'"
        );
    };
    gboolean ok = FALSE;
    gchar *  str = nullptr;
    g_variant_get( reply.get(), format, & ok, & str );
    if ( ! ok ) {
        ERR( str );
    };
    return str;
};

strings_t
input_source_manager_t::keyboard_layouts(
) {
    typedef std::unique_ptr< JsonParser, deleter_t< JsonParser > > json_parser_p;
    strings_t layouts;
    // TODO: Join all names into a string which could be parsed without Json?
    string_t  reply = _eval( R"(
        var layouts = [];
        const input_sources = imports.ui.status.keyboard.getInputSourceManager().inputSources;
        for ( i in input_sources ) {
            const source = input_sources[ i ];
            layouts[ source.index ] = source.displayName;
        };
        layouts;
    )" );
    json_parser_p parser( json_parser_new() );
    GError * error = nullptr;
    json_parser_load_from_data( parser.get(), reply.c_str(), reply.size(), & error );
    CHECK( error );
    static char const * const message =
        "Can't parse Gnome Shell response: array of strings is expected but received";
    auto root = json_parser_get_root( parser.get() );
    if ( not JSON_NODE_HOLDS_ARRAY( root ) ) {
        ERR( message << " `" << reply << "'" );
    };
    auto array = json_node_get_array( root );
    auto len = json_array_get_length( array );
    for ( size_t i = 0; i < len; ++ i ) {
        auto elem = json_array_get_element( array, i );
        if ( not JSON_NODE_HOLDS_VALUE( elem ) ) {
            ERR( message << " `" << reply << "'" );
        };
        auto name = json_node_get_string( elem );
        if ( not name ) {
            ERR( message << " `" << reply << "'" );
        };
        layouts.push_back( name );
    };
    return layouts;
};

void
input_source_manager_t::activate_layout(
    uint_t  idx,
    bool    bell
) {
    /*
        If Gnome Shell is restarted, activating layout fails with error
        GDBus.Error:org.freedesktop.DBus.Error.ServiceUnknown. Resetting connection helps to
        restore normal functioning.
    */
    size_t attempts = 0;
    start:
    try {
        static char const * const play_bell =
            "global.play_theme_sound( 0, 'bell', \"Keyboard layout selected\", null );";
        // TODO: eval async?
        auto reply = _eval(
            STR(
                "const idx = " << idx << "; "
                "const sources = imports.ui.status.keyboard.getInputSourceManager().inputSources; "
                "if ( idx in sources ) {"
                    "sources[ idx ].activate( false );" <<
                    ( bell ? play_bell : "" ) <<
                    "true;"
                "} else {"
                    "false;"
                "};"
            )
        );
        if ( reply == "true" ) {
            // Ok.
        } else if ( reply == "false" ) {
            WRN( "There is no layout #" << ( idx + 1 ) << "." );
        } else {
            ERR( reply );
        };
    } catch ( error_t const & ex ) {
        if ( ex.matches( G_DBUS_ERROR, G_DBUS_ERROR_SERVICE_UNKNOWN ) and attempts < 3 ) {
            WRN( ex.what() );
            WRN( "Looks like Gnome Shell was restarted, resetting connection…" );
            _reset();
            ++ attempts;
            goto start;
        };
        throw;
    };
};

}; // namespace gnome

// end of file //
