/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/posix.cpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    POSIX wrappers implementation.
**/

#include "posix.hpp"

#include <errno.h>
#include <libgen.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

namespace posix {

string_t
get_env(
    string_t const & name
) {
    char const * value = getenv( name.c_str() );
    return value ? value : "";
}; // get_env

string_t
base_name(
    string_t const & path
) {
    std::vector< char > buffer( path.c_str(), path.c_str() + path.size() + 1 );
    char * name = ::basename( buffer.data() );
    return name;
}; // basename

string_t
syserrmsg(
    int error
) {
    char buffer[ 1000 ];
    return strerror_r( error, buffer, sizeof( buffer ) );
}; // syserrmsg

// -------------------------------------------------------------------------------------------------
// error_t
// -------------------------------------------------------------------------------------------------

error_t::error_t(
    string_t const &    what,
    int                 error
) :
    std::runtime_error( STR( what << ": Error " << error << ": " << syserrmsg( error ) ) )
{
}; // ctor

error_t::error_t(
    string_t const &    what
) :
    std::runtime_error( what )
{
}; // ctor

// =================================================================================================
// signal
// =================================================================================================

namespace signal {

// -------------------------------------------------------------------------------------------------
// set_t
// -------------------------------------------------------------------------------------------------

/**
    Constructs empty signal set.
**/
set_t::set_t(
) {
    clear();
}; // ctor

/**
    Clears signal set. After invoking this methos signal set will be empty. Internally the method
    calls `sigemptyset` and checks error.
**/
void
set_t::clear(
) {
    int error = sigemptyset( & rep );
    if ( error ) {
        error = errno;
        ERR( "Clearing signal set failed", error );
    }; // if
}; // clear

/**
    Adds a signal to the signal set. Internally the method calls `sigaddset` and checks error.
**/
set_t &
set_t::operator +=(
    int signo
) {
    int error = sigaddset( & rep, signo );
    if ( error ) {
        error = errno;
        ERR( "Adding signal to signal set failed", error );
    }; // if
    return * this;
}; // +=

set_t
mask(
    int             how,
    set_t const &   set
) {
    set_t old;
    int error = pthread_sigmask( how, & set.rep, & old.rep );
    if ( error ) {
        ERR( "Setting signal mask failed", error );
    }; // if
    return old;
}; // mask

/**

**/
int
wait(
    set_t const &   set
) {
    TRACE();
    int signo = 0;
    int error = sigwait( & set.rep, & signo );
    if ( error ) {
        ERR( "Waiting for signal failed", error );
    }; // if
    return signo;
}; // wait

void
kill(
    pid_t pid,
    int signal
) {
    int error = ::kill( pid, signal );
    if ( error ) {
        error = errno;
        ERR( "Sending signal failed", error );
    }; // if
}; // kill

}; // namespace signal

// -------------------------------------------------------------------------------------------------
// thread_t
// -------------------------------------------------------------------------------------------------

static thread_local uint_t _thread_no = 0;

thread_t::thread_t(
) :
    OBJECT_T(),
    _state( inited )
{
    TRACE();
    DBG( "inited" );
}; // ctor

thread_t::~thread_t(
) {
    TRACE();
    CATCH_ALL(
        if ( _state == started ) {
            join();
        }; // if
    );
}; // dtor

void
thread_t::start(
) {
    TRACE();
    pthread_t thread;
    int error = pthread_create( & thread, nullptr, _body, this );
    if ( error ) {
        ERR( "Starting thread failed", error );
    }; // if
    DBG( "started" );
    _state  = started;
    _handle = thread;
}; // start

void
thread_t::join(
) {
    TRACE();
    switch ( _state ) {
        case inited : {
            DBG( "not started, nothing to do" );
        } break;
        case started : {
            void * ret = nullptr;
            int error = pthread_join( _handle, & ret );
            if ( error ) {
                ERR( "Joining thread failed", error );
            } else {
                DBG( "joined" );
            }; // if
            _state = joined;
            if ( not _what.empty() ) {
                DBG( "Thread: rethrowing exception occurred in thread body: " << _what );
                ERR( _what );
            }; // if
        } break;
        case joined : {
            DBG( "already joined, nothing to do" );
        } break;
        default : {
            ERR( "Oops" );
        } break;
    }; // switch
}; // join

uint_t
thread_t::no(
) {
    return _thread_no;
}; // id

void *
thread_t::_body(
    void * arg
) {
    thread_t * thread = reinterpret_cast< thread_t * >( arg );
    #if defined( DEBUG )
        _thread_no = thread->_no_();
    #endif
    TRACE1( thread );
    try {
        DBG1( thread, "really started" );
        thread->body();
    } catch ( std::exception const & ex ) {
        DBG1( thread, "exception caught: " << ex.what() );
        thread->_what = ex.what();
    } catch ( ... ) {
        DBG1( thread, "unknown exception caught" );
        thread->_what = "Unknown exception occurred";
    }; // try
    return nullptr;
}; // _body

}; // namespace posix

// end of file //
