/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/strings.hpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    String manipulating utilities interface.
**/

#ifndef _STRING_HPP_
#define _STRING_HPP_

#include "config.h"

#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

/**
    Build a string.

    Example:

    @code {.cpp}
    string_t str = STR( "string literal" << number << whatever_else );
    @endcode
**/
#define STR( _ARGS_ ) \
    reinterpret_cast< std::ostringstream & >( std::ostringstream() << _ARGS_ ).str()

/**
    Shorter alias for frequently used type.
    Yeah, `string_t` is shorter than `std::string`. :-)
**/
typedef std::string     string_t;

/** Shorter alias for frequently used type. **/
typedef std::vector< string_t >     strings_t;

/** Shorter alias for frequently used constant. **/
auto const npos = string_t::npos;

strings_t   split( char chr, string_t const & str, size_t limit = 0 );
size_t      find( string_t const & needle, strings_t const & haystack );
string_t    left( string_t const & str, size_t len );
string_t    right( string_t const & str, size_t len );
bool        like_integer( string_t const & str );
string_t    uc( string_t const & str );
string_t    chop( string_t const & str, string_t const & sfx );

class val_error_t : public std::runtime_error {
    public:
        explicit val_error_t( string_t const & message ) : std::runtime_error( message ) {};
};

template<
    typename type_t
>
type_t
val(
    string_t const & str
) {
    std::istringstream istr( str );
    type_t value = 0;
    istr >> std::noskipws >> value;
    if ( istr.fail() or istr.bad() ) {
        throw val_error_t( "Out of range." );
    }; // if
    // Now try to read one more character. We expect failure and EOF.
    char ch;
    istr >> std::noskipws >> ch;
    if ( not ( istr.fail() and istr.eof() ) ) {
        throw val_error_t( STR( "Invalid character `" << ch << "'." ) );
    }; // if
    return value;
}; // val

#endif // _STRING_HPP_

// end of file //
