/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/tapper.cpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

#include "tapper.hpp"

#include <limits>

#include <cassert>

#include <X11/Xproto.h>     // xEvent

#include "posix.hpp"

// -------------------------------------------------------------------------------------------------
// tapper_t::interceptor_t
// -------------------------------------------------------------------------------------------------

class tapper_t::interceptor_t : public x::record::context_t {

    public :

        explicit interceptor_t( tapper_t & _master );

    protected :

        virtual void on_intercept( XRecordInterceptData * data );

    private :

        /**
            Keyboard state. It maps key codes to boolean values. `true` means the
            corresponding key is pressed now, `false` means the key is not pressed.
        **/
        typedef std::vector< bool > kb_state_t;

        /**
            Reference to the owner. When interceptor detects a tap, master's `on_tap`
            method will be called.
        **/
        tapper_t & master;

        /** Keyboard state. **/
        kb_state_t state;

        /**
            Number of currently pressed keys. The number of currently pressed keys can be
            achieved from keyboard `state`, but scanning the `state` map is rather slow.
            Maintaining number of pressed keys is simple and fast operation.
        **/
        uint_t pressed_keys = 0;

        /**
            Key code of the last pressed key or -1. Used to detect taps: if code of
            released key equals to code of the last pressed key, it could be a tap.

            X.h defines `KeyCode` type as `unsigned char`, so -1 is not a valid key code.
            -1 means the last pressed key is not known (e. g. initially) or the last key
            press event cannot (or should not) be considered an initial phase of tap (e. g.
            when autorepeating occurs).
        **/
        int last_pressed_key = -1;
        // Make sure KeyCode is unsigned, and so, -1 is not a valid key code.
        static_assert( not std::numeric_limits< KeyCode >::is_signed, "oops" );

        /**
            Time of the last key press event, which code was saved in `last_pressed_key`.
            Meaningful only if `last_pressed_key` is not -1.
        **/
        Time pressed_at;

}; // class tapper_t::interceptor_t

// -------------------------------------------------------------------------------------------------
// tapper_t
// -------------------------------------------------------------------------------------------------

tapper_t::tapper_t(
    x::kb_t & kb
) :
    OBJECT_T(),
    _kb( kb ),
    interceptor( new interceptor_t( * this ) )
{
    autorepeat_rate = _kb.autorepeat_rate();
}; // ctor

tapper_t::~tapper_t(
) {
}; // dtor

void
tapper_t::start(
) {
    TRACE();
    interceptor->enable();
}; // start

void
tapper_t::stop(
) {
    TRACE();
    interceptor->disable();
}; // stop

// -------------------------------------------------------------------------------------------------
// tapper_t::interceptor_t
// -------------------------------------------------------------------------------------------------

tapper_t::interceptor_t::interceptor_t(
    tapper_t &  _master
) :
    master( _master ),
    state( std::numeric_limits< KeyCode >::max() )
{
}; // ctor

void
tapper_t::interceptor_t::on_intercept(
    XRecordInterceptData * data
) {
    TRACE();
    if ( data->category == XRecordFromServer ) {
        /*
            Decoding data described in:
            http://www.x.org/releases/X11R7.7/doc/xproto/x11protocol.html#events:input
        */
        if ( data->data_len * 4 < sizeof( xEvent ) ) {
            // `data_len` is size of data in 4-byte units.
            //~ ERR( "Oops" );
        }; // if
        xEvent * event = reinterpret_cast< xEvent * >( data->data );
        // My experiments show that data->server_time == event->u.keyButtonPointer.time.
        switch ( event->u.u.type ) {
            case KeyPress   :
            case KeyRelease : {
                bool    pressed  = ( event->u.u.type == KeyPress );
                int     key_code = event->u.u.detail;
                Time    time     = event->u.keyButtonPointer.time;
                //~ int     state    = event->u.keyButtonPointer.state;
                if ( pressed ) {
                    if ( state[ key_code ] ) {
                        // Look like key press is autorepeating.
                        if ( last_pressed_key > 0 ) {
                            // BTW, we are having autorepeat timeout for free...
                            // TODO: Enable it. Print a log message.
                            //~ master.autorepeat_rate.timeout = time - pressed_at;
                        }; // if
                        /*
                            No need in updating keyboard state -- we already know the key is
                            pressed. Number of pressed keys is not changed too. But let us reset
                            the last pressed key code and time -- if autorepeating takes place
                            this is not a "tap".
                        */
                        last_pressed_key = -1;
                        DBG( key_code << "↡ " << pressed_keys );
                    } else {
                        // Update keyboard state.
                        state[ key_code ] = true;
                        pressed_keys += 1;
                        // And remember the last pressed key code and time when it was pressed.
                        last_pressed_key = key_code;
                        pressed_at       = time;
                        DBG( "k" << key_code << "↓ " << pressed_keys );
                    }; // if
                } else {
                    /*
                        This check is a must because unconditional decreasing `pressed_keys` may
                        cause wrong result. For example, if the program started from the command
                        line, the first received event will likely be releasing of Enter key.
                    */
                    if ( state[ key_code ] ) {
                        state[ key_code ] = false;
                        assert( pressed_keys > 0 );
                        pressed_keys -= 1;
                    }; // if
                    if (
                        key_code == last_pressed_key
                        and pressed_keys == 0
                        and time - pressed_at <= master.autorepeat_rate.timeout
                    ) {
                        master.on_tap( key_code );
                    }; // if
                    last_pressed_key = -1;
                    DBG( "k" << key_code << "↑ " << pressed_keys );
                }; // if
            } break;
            case ButtonPress   :
            case ButtonRelease : {
                bool    pressed  = ( event->u.u.type == ButtonPress );
                int     button   = event->u.u.detail;
                //~ Time    time     = event->u.keyButtonPointer.time;
                static_cast< void >( button );      // Avoid cppcheck warning "buton never used".
                last_pressed_key = -1;
                if ( pressed ) {
                    DBG( "b" << button << "↓" );
                } else {
                    DBG( "b" << button << "↑" );
                }; // if
            } break;
        }; // switch
    }; // if
}; // on_intercept

// end of file //
