/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/x.cpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    X Window System wrappers implementation.
**/

#include "x.hpp"

#include <unistd.h>

namespace x {

static XErrorHandler _standard_error_handler = nullptr;

// Some of `XErrorEvent` fiedls are of `unsigned char` type, and printed as characters. It is
// undesired behavior, let us fix it and pring them as integer numbers.
static
std::ostream &
operator <<(
    std::ostream &  stream,
    unsigned char   ch
) {
    stream << uint_t( ch );
    return stream;
}; // operator <<

static
int
error_handler(
    Display *       display,
    XErrorEvent *   event
) {
    #if 1
        char error_text[ 1000 ];
        XGetErrorText( event->display, event->error_code, error_text, sizeof( error_text ) );
        // TODO: How to decode opcode?
        // If I could decode opcode, I would not call standard error handler.
        ERR(
            "Display " << display << " reports error:\n"
            "    Display connection : " << XConnectionNumber( event->display ) << "\n"
            "    Request number     : " << event->serial << "\n"
            "    Opcode             : " << event->request_code << "." << event->minor_code << "\n"
            "    Error              : " << event->error_code << " (" << error_text << ")\n"
        );
    #endif
    #if 1
        if ( _standard_error_handler ) {
            _standard_error_handler( display, event );
        }; // if
    #endif
    return 0;
}; // error_handler

void
init(
) {
    Status ok = XInitThreads();
    if ( not ok ) {
        ERR( "Initializing X failed" );
    }; // if
    _standard_error_handler = XSetErrorHandler( error_handler );
    DBG( "error handler installed" );
}; // init

// -------------------------------------------------------------------------------------------------
// error_t
// -------------------------------------------------------------------------------------------------

error_t::error_t(
    string_t const & msg
) :
    std::runtime_error( msg )
{
}; // ctor

// -------------------------------------------------------------------------------------------------
// display_t
// -------------------------------------------------------------------------------------------------

display_t::display_t(
) :
    OBJECT_T()
{
    TRACE();
    rep = XOpenDisplay( nullptr );
    if ( not rep ) {
        ERR( "Opening display `" << XDisplayName( nullptr ) << "' failed" );
    }; // if
    DBG(
        "opened, "
            << "name `" << XDisplayString( rep ) << "', "
            << "connection " << connection()
    );
}; // ctor

display_t::~display_t(
) {
    TRACE();
    CATCH_ALL(
        if ( rep ) {
            Status ok = XCloseDisplay( rep );
            // XCloseDisplay always returns 0. WTF??
            ok = ok || 1;       // Simple assignment causes cppcheck error.
            if ( not ok ) {
                ERR( "Closing display failed" );
            }; // if
            rep = nullptr;
            DBG( "closed" );
        }; // if
    );
}; // dtor

int
display_t::connection(
) const {
    return XConnectionNumber( rep );
}; // connection

void
display_t::flush(
) {
    Status ok = XFlush( rep );
    if ( not ok ) {
        ERR( "Flushing display failed" );
    }; // if
}; // flush

void
display_t::sync(
    bool discard
) {
    Status ok = XSync( rep, discard );
    if ( not ok ) {
        ERR( "Synchronizing display failed" );
    }; // if
}; // sync

// -------------------------------------------------------------------------------------------------
// atom_t
// -------------------------------------------------------------------------------------------------

atom_t::atom_t(
    display_t & display,
    Atom        atom
) :
    OBJECT_T(),
    _display( display.rep ),
    rep( atom )
{
}; // ctor

atom_t::atom_t(
    Display *   display,
    Atom        atom
) :
    OBJECT_T(),
    _display( display ),
    rep( atom )
{
}; // ctor

string_t
atom_t::name(
) {
    char * name = XGetAtomName( _display, rep );
    // cppcheck-suppress nullPointerRedundantCheck
    if ( not name ) {
        ERR( "Getting atom name failed" );
    }; // if
    string_t res = name;
    XFree( name );
    return res;
}; // name

// =================================================================================================
// record
// =================================================================================================

/*
    Primary documentation:
    http://www.x.org/releases/X11R7.7/doc/libXtst/recordlib.html
*/

namespace record {

// -------------------------------------------------------------------------------------------------
// intercept_data_t
// -------------------------------------------------------------------------------------------------

intercept_data_t::intercept_data_t(
    XRecordInterceptData * _data
) :
    OBJECT_T(),
    data( _data )
{
}; // ctor

intercept_data_t::~intercept_data_t(
) {
    if ( data ) {
        XRecordFreeData( data ); // void function
        data = nullptr;
    }; // if
}; // dtor

// -------------------------------------------------------------------------------------------------
// range_t
// -------------------------------------------------------------------------------------------------

range_t::range_t(
    unsigned char first,
    unsigned char last
) :
    OBJECT_T()
{
    rep = XRecordAllocRange();
    if ( not rep ) {
        ERR( "Allocating record range failed" );
    }; // if
    rep->device_events.first = first;
    rep->device_events.last  = last;
    DBG( "allocated" );
}; // ctor

range_t::~range_t() {
    CATCH_ALL(
        if ( rep ) {
            Status ok = XFree( rep );
            if ( not ok ) {
                ERR( "Freeing record range failed" );
            }; // if
            DBG( "freed" );
            rep = nullptr;
        }; // if
    );
}; // dtor

// -------------------------------------------------------------------------------------------------
// context_t
// -------------------------------------------------------------------------------------------------

context_t::context_t(
    int                 datum_flags,
    XRecordClientSpec   client
) :
    OBJECT_T(),
    _thread( _data )
{

    TRACE();

    // TODO: Move version check into a separate function?
    // TODO: Do we need to check version?
    int major = 0;
    int minor = 0;
    Status ok = XRecordQueryVersion( _ctrl.rep, & major, & minor );
    if ( not ok ) {
        ERR( "Querying X record extension version failed" );
    }; // if
    DBG( "X record extension version " << major << "." << minor );

    range_t         keys(       KeyPress,    KeyRelease );
    range_t         buttons( ButtonPress, ButtonRelease );
    XRecordRange *  ranges[ 2 ] = { keys.rep, buttons.rep };
    _handle = XRecordCreateContext(
        _ctrl.rep,
            /*
                This request is typically executed by the recording client over its control
                connection to the X server.
                http://www.x.org/releases/X11R7.7/doc/libXtst/recordlib.html#XRecordCreateContext
            */
        datum_flags,
        & client,
        1,          // Number of clients.
        ranges,
        sizeof( ranges ) / sizeof( XRecordRange * )     // Number of ranges.
    );
    if ( _handle == None ) {
        ERR( "Creating record context failed" );
    }; // if
    DBG( "created" );

}; // ctor

context_t::~context_t(
) {
    TRACE();
    if ( _handle != None ) {
        CATCH_ALL(
            if ( _state == enabled ) {
                disable();
            }; // if
        );
        CATCH_ALL(
            Status ok = XRecordFreeContext( _ctrl.rep, _handle );
            if ( not ok ) {
                ERR( "Freeing record context failed" );
            }; // if
            _handle = None;
            DBG( "freed" );
        );
    };
}; // dtor

void
context_t::enable(
) {
    TRACE();
    _ctrl.sync();
    /*
        Sync is a must. Otherwise XRecordEnableContextAsync will likely fail with error
        XRecordBadContext.
    */
    Status ok = XRecordEnableContextAsync(
        _data.rep,
        _handle,
        interceptor,
        reinterpret_cast< XPointer >( this )
    );
    if ( not ok ) {
        ERR( "Enabling record context failed" );
    }; // if
    /*
        My experiments show interceptor is called once with event XRecordStartOfData from within
        XRecordEnableContextAsync, so context should be enabled now.
    */
    if ( _state != enabled ) {
        ERR( "Record context is not really enabled" );
    }; // if
    _thread.start();
}; // enable

void
context_t::disable(
) {
    TRACE();
    switch ( _state ) {
        case inited : {
            DBG( "not enabled, nothing do do" );
        } break;
        case enabled : {
            /*
                To direct the X server to halt the reporting of recorded protocol, the
                program executes XRecordDisableContext typically over its **control**
                connection to the X server.

                The XRecordDisableContext function disables context, stopping all
                recording over its data connection. Any complete protocol elements for
                context that were buffered in the server will be sent to the recording
                client rather than being discarded. If a program attempts to disable an
                XRecordContext that has not been enabled, no action will take place.

                http://www.x.org/releases/X11R7.7/doc/libXtst/recordlib.html#Disable_Context
            */
            Status ok = XRecordDisableContext( _ctrl.rep, _handle );
            if ( ok ) {
                DBG( "disabled" );
            } else {
                ERR( "Disabling record context failed" );
            }; // if
            _ctrl.sync();   // Sync is a must.
            _thread.join();
            if ( _state != disabled ) {
                ERR( "Record context is not really disabled" );
            }; // if
        } break;
        case disabled : {
            DBG( "already disabled, nothing do do" );
        } break;
        default : {
            ERR( "Oops" );
        } break;
    }; // switch
}; // disable

void
context_t::interceptor(
    XPointer                closure,
    XRecordInterceptData *  data
) {
    /*
        (This code is executed by (non-initial) thread.)
        Xlib is a C library, it does not work with exception. If an exception leaked from this code,
        Xlib leaves an internal mutex locked, and program will hang. So we have to catch all the
        exceptions here.
    */
    context_t * context = reinterpret_cast< context_t * >( closure );
    TRACE1( context );
    try {
        intercept_data_t intercept_data( data );
        switch ( data->category ) {
            case XRecordStartOfData : {
                // XServer sends such record at the very beginning.
                context->_state = enabled;
            } break;
            case XRecordEndOfData : {
                // XServer sends such record at the very end.
                context->_state = disabled;
            } break;
            case XRecordFromServer : {
                context->on_intercept( data );
            } break;
        }; // switch
    } catch ( std::exception const & ex ) {
        // No need to warn: We just save the message now, an exception will be thrown when initial
        // thread calls `join`.
        context->_thread._what = ex.what();
        CATCH_ALL1( context, posix::signal::kill( getpid(), SIGTERM ); );
    } catch ( ... ) {
        context->_thread._what = "Unknown exception occurred";
        CATCH_ALL1( context, posix::signal::kill( getpid(), SIGTERM ); );
    }; // try
}; // interceptor

context_t::thread_t::thread_t(
    display_t & dsp
) :
    _display( dsp )
{
}; // ctor

void
context_t::thread_t::body(
) {
    TRACE();
    _display.sync();
}; // body

}; // namespace record

// -------------------------------------------------------------------------------------------------
// kb
// -------------------------------------------------------------------------------------------------

/*
    http://www.x.org/releases/X11R7.7/doc/libX11/XKB/xkblib.html
*/

kb_t::desc_t::desc_t(
    kb_t &  kb,
    uint_t  which
) :
    OBJECT_T()
{
    rep = XkbGetKeyboard( kb.display().rep, which, XkbUseCoreKbd );
    if ( not rep ) {
        ERR( "Getting keyboard description failed" );
    }; // if
}; // ctor

kb_t::desc_t::~desc_t(
) {
    if ( rep ) {
        XkbFreeKeyboard( rep, 0, True );
    }; // if
}; // dtor

kb_t::desc_t::key_code_range_t
kb_t::desc_t::key_code_range(
) const {
    kb_t::desc_t::key_code_range_t range;
    range.min = rep->min_key_code;
    range.max = rep->max_key_code;
    return range;
}; // key_code_range_t

strings_t
kb_t::desc_t::key_names(
) const {
    strings_t key_names( rep->max_key_code + 1 );
    for ( int i = rep->min_key_code, end = rep->max_key_code; i <= end; ++ i ) {
        // Key name is a fixed-length string padded with zeros.
        // We need to trim trailing zeros.
        char const * name = rep->names->keys[ i ].name;
        int len = XkbKeyNameLength;
        while ( len > 0 and name[ len - 1 ] == 0 ) {
            -- len;
        }; // while
        key_names[ i ] = string_t( name, len );
    }; // for i
    return key_names;
}; // key_names

strings_t
kb_t::desc_t::group_names(
) const {
    strings_t group_names( layout_max );
    for ( size_t i = 0, end = layout_max; i < end; ++ i ) {
        if ( rep->names->groups[ i ] != None ) {
            group_names[ i ] = atom_t( rep->dpy, rep->names->groups[ i ] ).name();
        }; // if
    }; // for i
    return group_names;
}; // group_names

kb_t::kb_t(
    display_t & display
) :
    OBJECT_T(),
    _display( display )
{
    {
        int major = XkbMajorVersion;
        int minor = XkbMinorVersion;
        Bool ok = XkbLibraryVersion( & major, & minor );
        if ( not ok ) {
            ERR(
                "X keyboard extension: "
                    << "Incompatible client library version " << major << "." << minor << "; "
                    << "required version " << XkbMajorVersion << "." << XkbMinorVersion
            );
        }; // if
        DBG( "X keyboard extension: client library version " << major << "." << minor );
    }
    {
        int major = XkbMajorVersion;
        int minor = XkbMinorVersion;
        int dummy1, dummy2, dummy3;
        Bool ok = XkbQueryExtension( _display.rep, & dummy1, & dummy2, & dummy3, & major, & minor );
        if ( not ok ) {
            ERR(
                "X keyboard extension: "
                "Incompatible server extension " << major << "." << minor << "; "
                    << "required version " << XkbMajorVersion << "." << XkbMinorVersion
            );
        }; // if
        DBG( "X keyboard extension: server extension version " << major << "." << minor );
    }
}; // ctor

display_t &
kb_t::display(
) {
    return _display;
}; // display

kb_t::autorepeat_rate_t
kb_t::autorepeat_rate(
) {
    autorepeat_rate_t rate;
    Bool ok = XkbGetAutoRepeatRate( _display.rep, XkbUseCoreKbd, & rate.timeout, & rate.interval );
    if ( not ok ) {
        ERR( "Getting autorepeat rate failed" );
    }; // if
    DBG( "autorepeat rate: timeout = " << rate.timeout << ", interval = " << rate.interval );
    return rate;
}; // autorepeat_rate

void
kb_t::lock_group(
    uint_t  group
) {
    TRACE();
    Bool ok = XkbLockGroup( _display.rep, XkbUseCoreKbd, group );
    if ( not ok ) {
        ERR( "Locking group " << ( group + 1 ) << " failed" );
    }; // if
    DBG( "group " << ( group + 1 ) << " locked" );
}; // lock_group

void
kb_t::bell(
) {
    Bool ok = XkbBell( _display.rep, None, 100, None );
    if ( not ok ) {
        ERR( "Bell failed" );
    }; // if
}; // bell

kb_t::desc_p
kb_t::desc(
    uint_t  which
) {
    return desc_p( new desc_t( * this, which ) );
}; // desc

}; // namespace x

// end of file //
