/*
    ---------------------------------------------------------------------- copyright and license ---

    file: src/x.hpp

    Copyright © 2014, 2016—2018 Van de Bugger.

    This file is a part of Tapper.

    Tapper is free software: you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    Tapper is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along with Tapper.  If not,
    see <http://www.gnu.org/licenses/>.

    SPDX-License-Identifier: GPL-3.0-or-later

    ---------------------------------------------------------------------- copyright and license ---
*/

/**
    @file
    X Window System wrappers interface.
**/

#ifndef _X_HPP_
#define _X_HPP_

#include "base.hpp"

#include "posix.hpp"

#include <stdexcept>        // std::runtime_error

#include <X11/Xlib.h>
#include <X11/XKBlib.h>
#include <X11/extensions/record.h>
#include <X11/extensions/XKB.h>

namespace x {

    uint_t const layout_max = XkbNumKbdGroups;

    void init();

    // ---------------------------------------------------------------------------------------------
    // error_t
    // ---------------------------------------------------------------------------------------------

    class error_t : public std::runtime_error {
        public :
            explicit error_t( string_t const & msg );
    }; // class error_t

    // ---------------------------------------------------------------------------------------------
    // display_t
    // ---------------------------------------------------------------------------------------------

    class display_t : public object_t {

        public :

            explicit display_t();
            virtual ~display_t();

            int connection() const;

            void flush();
            void sync( bool discard = false );

        public :

            Display * rep = nullptr;

    }; // class display_t

    typedef ptr_t< display_t > display_p;

    // ---------------------------------------------------------------------------------------------
    // atom_t
    // ---------------------------------------------------------------------------------------------

    class atom_t : public object_t {

        public :

            explicit atom_t( display_t & display, Atom atom );
            explicit atom_t( Display * display, Atom atom );

            string_t name();

        private:

            Display *   _display = nullptr;
            Atom        rep;

    }; // atom_t

    namespace record {

        // -----------------------------------------------------------------------------------------
        // intercept_data_t
        // -----------------------------------------------------------------------------------------

        class intercept_data_t : public object_t {

            public :

                explicit intercept_data_t( XRecordInterceptData * data = nullptr );
                virtual ~intercept_data_t();

            public :

                XRecordInterceptData * data;

        }; // class intercept_data_t

        // -----------------------------------------------------------------------------------------
        // range_t
        // -----------------------------------------------------------------------------------------

        class range_t : public object_t {

            public :

                explicit range_t( unsigned char first, unsigned char last );
                virtual ~range_t();

            public :

                XRecordRange * rep = nullptr;

        }; // class range_t

        // -----------------------------------------------------------------------------------------
        // context_t
        // -----------------------------------------------------------------------------------------

        class context_t : public object_t {

            public :

                explicit context_t(
                    int               datum_flags = 0,
                    XRecordClientSpec client      = XRecordAllClients
                );
                virtual ~context_t();

                void enable();
                void disable();

            protected :

                virtual void on_intercept( XRecordInterceptData * data ) = 0;

            private :

                static void interceptor( XPointer closure, XRecordInterceptData * data );

                class thread_t : public posix::thread_t {
                    friend class context_t;
                    public :
                        explicit thread_t( display_t & dsp );
                    protected :
                        virtual void body();
                    private :
                        display_t & _display;
                }; // thread_t

            private :

                enum state_t {
                    inited,
                    enabled,
                    disabled
                }; // enum state_t

                state_t         _state  = inited;
                XRecordContext  _handle = None;
                display_t       _ctrl;
                display_t       _data;
                thread_t        _thread;

        }; // class context_t

    }; // namespace record

    // ---------------------------------------------------------------------------------------------
    // kb_t
    // ---------------------------------------------------------------------------------------------

    class kb_t : public object_t {

        public :

            struct autorepeat_rate_t {
                uint_t  timeout  = 0;
                uint_t  interval = 0;
            }; // struct autorepeat_rate_t

            /** Keyboard description. **/
            class desc_t : public object_t {

                friend class kb_t;

                private :

                    explicit desc_t( kb_t & kb, uint_t which );

                public :

                    struct key_code_range_t {
                        uint_t  min = 0;
                        uint_t  max = 0;
                    }; // struct key_code_range_t

                    virtual ~desc_t();

                    key_code_range_t key_code_range()   const;
                    strings_t        key_names()        const;
                    strings_t        group_names()      const;

                private :

                    XkbDescPtr  rep = nullptr;

            }; // class desc_t

            typedef ptr_t< desc_t > desc_p;

        public :

            explicit kb_t( display_t & display );

            display_t &       display();
            autorepeat_rate_t autorepeat_rate();
            void              lock_group( uint_t group );
            void              bell();
            desc_p            desc( uint_t which = XkbAllComponentsMask );

        private :

            display_t & _display;

    }; // class kb_t

    typedef ptr_t< kb_t >   kb_p;

}; // namespace x

#endif // _X_HPP_

// end of file //
